(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)


(* core.sml: A package that integrates abstract syntax for
   lambda terms, higher-order pattern unification, substitutions,
   definition clauses, parser for lambda terms, and pretty printer
   for lambda terms and streams.
   Basically, it contains everything that is needed to build a prover.
*)

signature CORE =
sig
	structure Term : TERM
	structure Norm : NORM
	structure LamTree : LAMTREE
	structure Cls : CLAUSE
	structure Sub : SUBST
	structure Unif : HO_PATTERN_UNIFY
	structure UStream : USTREAM
	structure Lam : LAM
	structure ParInt : PARSERINT
	structure PPrint : PPRINT
	sharing Norm.Term = Term and Cls.Term = Term and Sub.Term = Term 
		and Term = LamTree.Term 
		and Unif.Term = Term and Lam.Term = Term and ParInt.Term = Term
	sharing Lam.Cls = Cls and ParInt.Cls = Cls and PPrint.Term = Term
	sharing Lam.ParInt = ParInt
	sharing PPrint.Sub = Sub
	sharing PPrint.UStream = UStream

	exception InvalidGoal
	
	(* some wrapper functions *)	
	val unify : Term.term -> Term.term -> unit
	val norm  : Term.term -> Term.term
	val load : string -> Cls.clause list

	(* debugging options *)
	val set_debug : bool -> unit
	val is_debug : unit -> bool

	(* some shared global values *)
	val path : string ref
	(* val const_level : int ref  *)
        (* current constant level, i.e., timestamp for (eigen)variables *)
	val eq : Term.term 
	val judg : Term.term 
	val set_path : string -> unit

	(* some helper functions for raising, etc *)
	val make_judgments : int -> Term.term list -> Term.term list
	val new_dBs : int -> Term.term list
	val new_eigenvars : int -> int -> Term.term list
	val new_logicvars : int -> int -> Term.term list
	val raise_eigenvars : int -> int -> int -> Term.term -> Term.term
	val raise_logicvars : int -> int -> int -> Term.term -> Term.term

end

functor CoreFunc(
	structure Term : TERM
	structure Norm : NORM
	structure LamTree : LAMTREE
	structure Cls : CLAUSE
	structure Sub : SUBST
	structure Unif : HO_PATTERN_UNIFY
	structure UStream : USTREAM
	structure Lam : LAM
	structure ParInt : PARSERINT
	structure PPrint : PPRINT
	sharing Norm.Term = Term and Cls.Term = Term and Sub.Term = Term 
		and Term = LamTree.Term
		and Unif.Term = Term and Lam.Term = Term and ParInt.Term = Term
	sharing Lam.Cls = Cls and ParInt.Cls = Cls and PPrint.Term = Term
	sharing Lam.ParInt = ParInt
	sharing PPrint.Sub = Sub
	sharing PPrint.UStream = UStream
) : CORE =
struct
	structure Term = Term
	structure Norm = Norm
	structure LamTree = LamTree
	structure Cls = Cls
	structure Sub = Sub
	structure Unif = Unif
	structure UStream = UStream
	structure Lam = Lam
	structure ParInt = ParInt
	structure PPrint = PPrint

	structure T = Term
	
	exception InvalidGoal

	(* val const_level = ref 0 *)

	val eq = ref (Term.const("=", 0, Term.none))
	val judg = ref (Term.const("|> ", 0, Term.none))

	val unify = Unif.pattern_unify
	val norm = Norm.norm
	
	val debug = ref false

	fun is_debug() = !debug

	fun set_debug b = (debug := b)

	val path = ref ""
	fun set_path S = (path := S)

	fun load fname =
	let val instream = TextIO.openIn(fname)
		val L = Lam.load(instream)
	in
		(TextIO.closeIn(instream) ; L)
	end    
	

	(* 
   	 Local judgment 
	   x_1,..., x_n |> P 
	 is represented as the term
	   lam x_1 .. lam x_n. (|> P)
	*)  

	(* make_judgments n L
	   takes the list of formulas L and turns them into judgments 
	   n is the number of local variables.
	*)          

	fun make_judgments n nil = nil
	  | make_judgments n (t::l) = 
  		if (n = 0) then 
  			(ref (T.app(judg, [t]))) :: (make_judgments n l)
  		else 
  			(ref (T.lam(n,ref (T.app(judg,[t]))))) :: (make_judgments n l)

	(* create a list of deBruijn indices *)  			
	fun new_dBs 0 = nil
	  | new_dBs n = 
	     (ref (T.dB n)) :: (new_dBs (n-1))

	(* create new cells for eigenvariables *)     
	fun new_eigenvars ts 0 = nil
	  | new_eigenvars ts n = 
	    let val newvar = T.fresh_eigenvar (ts)
		in (newvar:: (new_eigenvars ts (n-1))) end    			

	(* create new logic variables *)	
	fun new_logicvars ts 0 = nil
	  | new_logicvars ts n = 
	    let val newvar = T.freshvar (ts)
		in (newvar:: (new_logicvars ts (n-1))) end    	
			  			
	(* raise eigenvariables 
	create m new eigenvariables and raise them 
	with respect to n (i.e., the number of locally bound variables).
	G' is an abstraction of the form (lam n. t)
	*)		
	fun raise_eigenvars ts n m G' =
		let val vars = new_eigenvars ts m
			val dBs = new_dBs n
			val ts = map (fn var => ref (T.app(var, dBs))) vars 
		in
			(ref (T.lam(n, ref (T.app(judg, [ref (T.app(G', ts))])))))
		end
  		
	(* raise logic variables *)			  
	fun raise_logicvars ts n m G' =
		let val vars = new_logicvars ts m
			val dBs = new_dBs n
			val ts = map (fn var => ref (T.app(var, dBs))) vars 
		in
			(ref (T.lam(n, 
			   	ref (T.app(judg, [ref (T.app(G', ts))]))
			   	))
			)
		end
		
end

