(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* Level 1 Prover: Horn-nabla plus implication, equality, universal *)

signature PROVER1 =
sig
	structure Core : CORE
	structure Prover0 : PROVER0
	sharing Prover0.Core = Core

	val prove : int  * Core.Cls.clause list * Core.Term.term list * Core.Sub.subst -> Core.Sub.subst Core.UStream.ustream
end


functor Prover1Func (
	structure Core : CORE
	structure Prover0 : PROVER0
	sharing Prover0.Core = Core
) : PROVER1 = 

struct

structure Core = Core
structure Prover0 = Prover0

open Core

structure T = Term
val hc_prove = Prover0.prove

fun body_and_sub(lev, [],_,_,_) = UStream.snil
  | body_and_sub(lev, clause::rest,n,atom,subst) =
    let val (head,body) = 
		Cls.fresh_clause(n, lev, clause) 
        val newhead = Norm.norm(ref (Term.lam(n, ref (Term.app(judg, [head])))))
		val newbody = Norm.norm(ref (Term.lam(n, ref (Term.app(judg, [body])))))
		val (A, newsub) =  Sub.copy_lv atom
    in
      (
        if (is_debug()) then 
		(
            print "Unifying goal : \n";
            PPrint.print_abs_term 0 A;
            print "\n with the (raised) clause :\n";
		  	PPrint.print_abs_term 0 newhead; 
		  	print " := ";
		  	PPrint.print_abs_term 0 newbody;
		  	print "\n"
	    )
		else ();
        (
	unify A newhead) ;  
      	if (is_debug()) then 
		(print "Unification succeeded.\n"; 
		 PPrint.print_substs newsub
		)
	else (); 
       	UStream.scons(([newbody], Sub.compose_sub subst newsub),
              (fn ()=>(body_and_sub(lev, rest,n,atom,subst))))
      )
         handle Unif.UniEx(E) =>
         	(
         	    if (is_debug()) then
         	       (print "Unification failed: "; print E; print "\n")
         	    else ();
         		body_and_sub(lev, rest,n,atom,subst)
         	)
         	|	Unif.ConstClash(_,_) =>
         	(
         	    if (is_debug()) then
         	       (print "Unification failed: constants clash"; print "\n")
         	    else ();
         		body_and_sub(lev, rest,n,atom,subst)         	
         	)
    end


fun makeConstTerm S = ref (Term.const(S,0,Term.none))

fun makeListTerm nil = makeConstTerm("nil")
  | makeListTerm (Cls.Clause(h,b) :: ts) = 
	let val d = LamTree.abstractDef(h,b)
	in
	  ref (Term.app(makeConstTerm("cons"), [d, makeListTerm ts]))
	end

fun extractTerms (ref (Term.const("nil",_,_))) = nil
  | extractTerms (ref (Term.app(ref (Term.const("cons",_,_)), (t::ts)))) =
  		(t:: (List.concat (map extractTerms ts)))
  | extractTerms t = [t]
    			    	
fun prove_atom (lev, prog, n, A, subst) =
	let val prog' = Cls.select_clauses(A, prog)
		val G = 
			(
				if (is_debug()) then 
					(PPrint.print_term A; print "\n")
				else ();
            	if (n = 0) then ref (Term.app(judg,[A]))
            	else ref (Term.lam(n, ref (Term.app(judg, [A]))))
            )
	in
        UStream.squash (UStream.maps (fn (body,s) => (prove(lev, prog,body,s)))
                    (body_and_sub(lev, prog',n,G,subst))) 
	end
	
and prove_eq (lev, prog, n, E, subst) =
    let val (E', newsub) = Sub.copy_lv E                
    	val E2 = Norm.hnorm E'
    	val (s,t) =
    	    case E2 of ref (Term.app(_, [u, v])) =>
    	            ( ref (Term.lam(n, u)), ref (Term.lam(n,v)))
    	       | _ => raise InvalidGoal
    in
		(unify s t; 
		 UStream.scons(Sub.compose_sub subst newsub,fn () => UStream.snil) )    
    	handle Unif.UniEx(E) => UStream.snil
    		|  Unif.ConstClash(_,_) => UStream.snil
    end
    
and prove_imp (lev, prog, n, A, B, subst) =
    let val A' = (Norm.norm A)
        val (G, L) = 
        	(
        		if (is_debug()) then
        			(
        				print "switching from level 1 to 0 : "; 
        			 	PPrint.print_abs_term n A'; 
        				print "\n"
        			)
        		else ();
        		Sub.copy_ev A'
        	)  (* make a copy of A, change eigenvars to logic vars *)
        val G' = ref (Term.lam(n,ref (Term.app(judg,[G])) ))
        val B' = ref (Term.lam(n,ref (Term.app(judg,[B])) ))
	(* call level 0 prover *)
        val S = hc_prove(lev, prog,[G'],L)
        (* Apply substitutions in the stream S to B', collect the resulting judgments
           in a list. Logic variables in the substitutions are changed back to
           eigenvariables. *)
        val Gs = UStream.mapsl (fn s => (Sub.sub_var_to_ev s; Sub.apply_sub s B')) S        	
    in
        (
        	if (is_debug()) then
        	(
        		print "Remaining goals: ";
        		PPrint.print_list Gs;
        		print "\n"
        	)
        	else ();
    		prove(lev, prog, Gs, subst)    	
    	)
    end    
    
and prove_disjunct (lev, prog, nil, subst) = UStream.snil
  | prove_disjunct (lev, prog, (G::rest), subst) =
      	UStream.scons(prove(lev, prog, [G], subst), 
      		fn () => prove_disjunct(lev, prog, rest, subst))

and prove_if (lev, prog, n, P, Q, R, subst) =
	let val S = prove_judgments (lev, prog, n, P, subst)
    in
    	if (S = UStream.snil) then
    		prove_judgments (lev, prog, n, R, subst)
    	else
    	(
    		let val G = if (n = 0) then ref (Term.app(judg, [Q]))
    					else ref (Term.lam(n, ref (Term.app(judg, [Q])) ))
    		in
				UStream.squash
				(UStream.maps
					(fn subst1 => 
						prove(lev, prog, [Sub.apply_sub subst1 G], subst1))
                    S 
                 )    			
    		end
    	)
    end

and prove_parse(lev, prog, n, S, P, subst) =
    let val L = Lam.loadFile(!path, S)
        val PL = makeListTerm(L)
        val G = ref (Term.app(eq, [P, PL]))
    in
	prove_judgments(lev, prog,n,G,subst)
    end
    
and prove_unparse(lev, prog, n, S, P, subst) =
	let val L = extractTerms(P)
		val R = map (LamTree.concreteSyntax) L
	in
		print "Terms extracted : \n";
		map (fn t => (PPrint.print_term t; print "\n")) R;		
		PPrint.writeFile(!path, S, R); 
		UStream.scons(subst,fn () => UStream.snil)
	end	

and prove_input(lev, prog, n, T, subst) =
	let val p = (Lam.getQuery())
		     handle _ => (NONE)     		  
    	  in
    	    case p of
    	       	(SOME (ParInt.ParseGoal t)) => 
    	       		let val goal = ref (Term.app(eq,[t,T]))
     				in 
 					prove_judgments(lev, prog, n, goal, subst)
     				end
     		  | _ => (print "Invalid term.\n"; UStream.snil)	
          end   		  
	
and prove_judgments (lev, prog, n, G, subst) =
	case G of
	  ref (T.app(ref (Term.const("&", 0, Term.none)), Gs)) => 
	  		(
	  		if (is_debug()) then
	  		 (
	  		 	print "prove conjunction : "; 
	  		 	PPrint.print_list Gs;
	  		 	print "\n"
	  		 )
	  		 else ();
	  		 prove (lev, prog, (make_judgments n Gs), subst)
	  		)
	  |  ref (T.app(ref (Term.const(";", 0, Term.none)), Gs)) => 
	  		(
	  		if (is_debug()) then
	  		 (
	  		 	print "prove disjunction : "; 
	  		 	PPrint.print_list Gs;
	  		 	print "\n"
	  		 )
	  		 else ();
	  		 UStream.squash(prove_disjunct(lev, prog, (make_judgments n Gs), subst))
	  		)	  		

      |  ref (T.app(ref (Term.const("if", 0, Term.none)), [P,Q,R])) => 	  
      		(
      			if (is_debug()) then
      			(
      				print "prove if-statement \n"
      			)
      			else ();
      			prove_if (lev, prog, n, P, Q, R, subst)
      		)
	  | ref (T.app(ref (Term.const("nabla", 0, Term.none)), [ref (T.lam(m,G'))])) =>
	  		if (m = 0) then (
	  			if (is_debug()) then print "Empty abstraction!\n" else ();
	  			raise InvalidGoal
	  			)
	  		else prove (lev, prog, (make_judgments (n + m) [G']), subst)
	  | ref (T.app(ref (Term.const("pi", 0, Term.none)), [G' as ref (T.lam(m,_))])) =>
	  		(
   	  		  if (m = 0) then (
   	  		  	if (is_debug()) then print "Empty abstraction!\n" else ();
   	  		  	raise InvalidGoal)
	  		  else 
			     let val newG = Norm.norm(raise_eigenvars (lev+1) n m G')
                 in
                    ( 
                    	if (is_debug()) then 
                        	(print "prove universal : ";
                                    PPrint.print_abs_term n newG;
                                    print "\n"
                    	)
                    	else ();
						prove (lev + 1, prog, [newG], subst)
                	)
                 end
	  		)
	  | ref (T.app(ref (Term.const("sigma", 0, Term.none)), [G' as ref (T.lam(m,_))])) =>
	  		if (m = 0) then (raise InvalidGoal)
	  		else prove (lev, prog, [(raise_logicvars lev n m G')], subst)
	  | ref (T.app(ref (Term.const("=>", 0, Term.none)), [A, B])) =>
	        (
	  		   prove_imp (lev, prog, n, A, B, subst)
	  		)
	  | E as ref (T.app(ref (Term.const("=", 0, Term.none)), [s,t])) =>
	        prove_eq (lev, prog, n, E, subst)
	  | ref (T.app(ref (Term.const("parse",_,_)), [ref (Term.const(S,_,_)), P])) =>
		prove_parse(lev, prog,n,S,P,subst)
	  | ref (T.app(ref (Term.const("unparse",_,_)), [ref (Term.const(S,_,_)), P])) =>
		prove_unparse(lev, prog,n,S,P,subst)	
	  | ref (T.app(ref (Term.const("input", _,_)), [T])) =>
		prove_input(lev, prog,n,T,subst)	
	  |	ref (T.app(ref (Term.const("print", 0, Term.none)), [s])) =>
		    ( 
		       (case s of ref (Term.const(str, _, _)) =>
		       		if (String.isPrefix "\"" str) then
		       		   if (String.size(str) > 2) then
		       		   		print (String.substring(str,1,String.size(str) - 2))
		       		   else ()
		       		else print str
                    | _ =>  PPrint.print_abs_term n s
		    	) ;
		        UStream.scons(subst,fn () => UStream.snil)
		    )	        
	  | ref (Term.const("true", 0, Term.none)) => UStream.scons(subst,fn () => UStream.snil)
	  | ref (Term.const("false", 0, Term.none)) => UStream.snil		        
	  | A as (ref (T.app(ref (T.const(_,_,_)), ts))) =>
	  		(
	  			if (is_debug()) then print "prove atom\n"
	  			else (); 
	  			prove_atom (lev, prog, n, A, subst)  
	  		)
	  | A as (ref (T.const(_,_,_))) =>
	  		(
	  			if (is_debug()) then print "prove atom\n"
	  			else (); 
	  			prove_atom (lev, prog, n, A, subst)  
	  		)	  			  		
	  | _ => (raise InvalidGoal)
  		        
and prove1 (lev, prog, A, subst) =
   let val (ref G) = Norm.norm A
   in
      case G of 
       	  T.app(judg,[G']) => prove_judgments (lev, prog, 0, G', subst)
      	| T.lam(n, ref (T.app(judg,[G']))) =>
      		prove_judgments (lev, prog, n, G', subst)
      	| _ => (PPrint.print_term (ref G) ; raise InvalidGoal)
   end        

and prove(lev, prog, [], subst) = UStream.scons(subst,fn () => UStream.snil)
  | prove(lev, prog, G::rest, subst) =
     UStream.squash
      (UStream.maps
        (fn subst1 =>
            prove(lev, prog,
                  map (fn a => Sub.apply_sub subst1 a) rest,
                  subst1))
       (prove1(lev, prog,G,subst)))


end (* of structure declaration *)
