(****************************************************************************)
(* An implementation of Higher-Order Pattern Unification in SML             *)
(* Copyright (C) 2004 Gopalan Nadathur and Natalie Linnell                  *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* Modified by Alwen Tiu, September 2004.                                   *)

(* 
term.sml: This module contains the data structure for lambda terms.
*)


signature TERM =
sig

datatype tag =  none  
			  | ev   (* indicating whether a variable/constant 
 							is an eigenvariable *)
 			  			
datatype rawterm =  const of string * int * tag
                 |  var of string * int  * tag
                 |  dB of int
                 |  lam of int * (rawterm ref)
                 |  app of (rawterm ref) * (rawterm ref) list
                 |  susp of (rawterm ref) * int * int * (envitem list)
                 |  ptr of (rawterm ref)
                 
and envitem = dum of int
            | bndg of (rawterm ref) * int
            

type term = (rawterm ref)
type env = (envitem list)

exception NonNormalTerm
exception NotValidTerm

val deref : term -> term
val nth : env*int -> envitem
val add_dummies : env -> int -> int -> env
val makelambda : int -> term -> term
val bind : term -> term -> unit
val freshvar : int -> term
val fresh_eigenvar : int -> term
val makeAbs : string * term -> term
val makeAbsVar : term * term -> term
val freevars : term list -> term list
val reset_counter : unit -> unit

val reset_name : unit -> unit
val getAbsName : unit -> string
val checkAndUpdateName : string -> unit
val set_prefixes : (string * string * string) -> unit
val get_prefixes : unit -> (string * string * string)
end


structure Term : TERM =
struct

structure Bs = Basics

val varcount = (ref 0)

datatype tag = none | ev 

datatype rawterm =  const of string * int * tag
                 |  var of string * int * tag
                 |  dB of int
                 |  lam of int * (rawterm ref)
                 |  app of (rawterm ref) * (rawterm ref) list
                 |  susp of (rawterm ref) * int * int * (envitem list)
                 |  ptr of (rawterm ref)
                 
and envitem = dum of int
            | bndg of (rawterm ref) * int

type term = (rawterm ref)

type env = (envitem list)

exception NonNormalTerm
exception NotValidTerm

val prefix1 = ref "h"
val prefix2 = ref "H"
val absName = ref "x"

fun set_prefixes(n1,n2,n3) = (prefix1 := n1; prefix2 := n2; absName := n3)
fun get_prefixes() = (!prefix1, !prefix2, !absName)

fun reset_name() = (prefix1 := "h"; prefix2 := "H"; absName := "x")

fun getAbsName() = (!absName)

fun updateName pre S =
    if (String.isPrefix (!pre) S) then
    (
	let val n1 = String.size (!pre)
	    val n2 = String.size S
	in
	    if (n1 = n2) then () 
	    else
		let val c = String.sub(S,n1)
		in
		   if ord(c) >= ord(#"0") andalso ord(c) <= ord(#"9") then
			pre := !pre ^ (Int.toString((ord(c) - ord(#"0") + 1) mod 10))
                   else pre := !pre ^ "0"
		end
	end
     )
     else ()        

fun checkAndUpdateName S = (updateName prefix1 S; updateName prefix2 S; updateName absName S)



fun reset_counter() = (varcount := 0)

fun deref (ref(ptr(t))) = deref(t)
  | deref term = term

fun nth(x::l,1) = x
  | nth(x::l,n) = nth(l,n-1)

fun add_dummies e 0 _ = e
  | add_dummies e n nl = 
       let val n' = (n-1) 
       in ((dum (nl+n')) :: (add_dummies e n' nl))
       end

(* Construct an abstraction from the no of abstractions and the body *)
fun makelambda 0 t = t
  | makelambda n (ref (lam(n',t))) = (ref (lam (n+n',t)))
  | makelambda n t = (ref (lam(n,t)))


(* Binding a variable to a term. The *contents* of the cell the
   variable is a reference to must be updated. Also the variable must
   not be made a reference to itself. This can be changed to mimic the
   Prolog representation of bound variables but then deref will have to
   work differently. This is the place to introduce trailing.
*)
fun bind v t = if (v = t) then () else (v := ptr(t))


(* Generating a fresh variable with a given time stamp; the use of ref
   ensures uniqueness. We should attach useful names as well, but this 
   will do for the moment. 
*)
fun freshvar ts2 = 
      let val name = concat([!prefix2, Int.toString(!varcount)])
      in
	varcount := !varcount + 1 ;
        ref (var (name,ts2, none))
      end

fun fresh_eigenvar ts =
	let val name = concat([!prefix1, Int.toString(!varcount)])
	in
		(varcount := !varcount + 1;
		 ref (const (name, ts, ev)))	 
	end         

   
fun makeAbs(id, t) = 
  let fun abs n (s as ref (var(id',i,f))) = 
  				if (id = id') then ref (dB n)
  				else s
  		| abs n (s as ref (const(id',i,f))) = 
  				if (id = id') then ref (dB n)
  				else s
  		| abs n (s as ref (dB i)) = s 
  		| abs n (ref (app(h, ts))) = 
  		     ref (app(abs n h, map (abs n) ts))
  		| abs n (ref (lam(m, s))) =
  		     ref (lam(m, abs (n+m) s))
        | abs n _ = raise NonNormalTerm
  in
  	ref (lam(1, abs 1 t))
  end

fun makeAbsVar(v, t) = 
  let fun abs n (s as ref (var _)) = 
  		  	if (v = s) then ref (dB n)
  		  	else s
  		| abs n (s as ref (const _ )) = 
  			if (v = s) then ref (dB n)
  			else s
  		| abs n (s as ref (dB i)) = s 
  		| abs n (ref (app(h, ts))) = 
  		     ref (app(abs n h, map (abs n) ts))
  		| abs n (ref (lam(m, s))) =
  		     ref (lam(m, abs (n+m) s))
        | abs n _ = raise NonNormalTerm
  in
  	ref (lam(1, abs 1 t))
  end

fun fv1 (t as (ref (var(_,_,_)))) L =
	if (Bs.member(t,L)) then L else (t::L)
  | fv1 (ref (app(h, ts))) L =
        fv2 ts (fv1 h L)
  | fv1 (ref (lam(n,t))) L =
        fv1 t L
  | fv1 t L = L
and fv2 nil L = L
  | fv2 (x::ts) L = fv2 ts (fv1 x L)

fun freevars ts = fv2 ts nil
    
end (* structure *)
